        program cvmbed
        implicit double precision(a-h,o-z)
        parameter(jdmax=30,nxmax=550)
        dimension xt(nxmax),xdat(nxmax),xlag(nxmax,jdmax),hvals(20)
        character*30 fname,outname
        common /xdata/ xt,xlag,cgcv,jd,nx,m,ipr,jw
	external rms

c----------- Read in parameters
        open(11,file='cvmbed.par',status='old')
        read(11,10) fname
        read(11,10) outname
10      format(a30)
        read(11,*) nx,jt1,ltr,ifreq,jd1,jd2,l1,l2,
     *       jtp1,jtp2,cgcv,ipr,jw
        close(11)
        csav=cgcv
        if (jw.lt.0) jw=0

c------------ read in data series; set nx=total series length
       call getx(fname,xdat,nxmax,nx,ifreq,ltr)

c---------- Set up the output files
        open(12,file=outname,status='unknown')
        write(12,110) fname,ltr
110     format(1x,'@ ',a30,'Output from CVEMBED.FOR',3x,'ltr=',i2)
        write(12,*) 'Data or transformed data scaled to Std Dev ==1'
        write(12,120) nx,jt1
120     format(1x,'max n= ',i6,' data pts, window length= ',i5)
        write(12,125) ifreq,cgcv,jw
125     format(1x, 'ifreq= ',i4, '  Cgcv= ',f5.2,' W= ',i4)
        write(12,*) 'output tp, l, d, CV_opt, h_opt, RMS_opt @'
        close(12)

c------------------- Scale so SD of values to predict == 1
       sd=sdev(xdat,jt1+1,nx)
       call dscal(nx,1./sd,xdat,1)
       xbar=0.
       do 140 j=1,nx
         xbar=xbar+xdat(j)
140    continue
       xbar=xbar/dfloat(nx)
       do 150 j=1,nx
         xdat(j)=xdat(j)-xbar
150    continue

c------------------ START MAIN LOOP on prediction time, time-delay, #lags
       do 520 jtp=jtp1,jtp2
       do 510 ldelay=l1,l2
          do 500 jd=jd1,jd2

          maxback=jtp+(jd-1)*ldelay
          if(jt1.lt.maxback) goto 500

          call recontp(xt,xdat,xlag,nxmax,jdmax,nx,ldelay,jd,jt1,
     *        m,jtp)

c
c try to bracket a minimum of CV(h)
c
	  cvmin=100.
	  imax=13
	  do 400 i=1,imax
		hi=10.**( -3.0+ 0.5*dfloat(i-1))
		hvals(i)=hi
		cvi=rms(hi)
		if (cvi.lt.cvmin) then
			cvmin=cvi
			imin=i
		endif
400	  continue

c
c if bracketing failed, use endpoint as optimal bandwidth
c
	  if (imin.eq.1) then
		hcv=hvals(1)
		cvfn=cvmin	
	  elseif (imin.eq.imax) then
		hcv=hvals(imax)
		cvfn=cvmin
	  else
c
c if bracketing succeeded, find the minimum
c
	  	tol=.0001
		ax=hvals(imin-1)
		bx=hvals(imin)
		cx=hvals(imin+1)
	   	cvfn=ffmin(ax,bx,cx,rms,tol,hcv)
	  endif
c
c save CV(h_opt)
c
          cgcv=1.
          rmscv=rms(hcv)
          cgcv=csav
	  open(12,file=outname,status='old',access='append') 
          write(12,210) jtp,ldelay,jd,cvfn,hcv,rmscv
	  close(12)
210       format(1x,3i3,3(2x,f9.4))

500    continue
510    continue
520    continue
c---------------------- END MAIN LOOP
       close(12)
       stop
       end

c------------------------------------------------------------------------
c        Leave-one-out kernel regression estimates
c------------------------------------------------------------------------
      subroutine kregfhat(h,n,m,xk,y,npts,nvar,f,params,jw)

c   Computes leave-one-out kernel regression estimates at each data point,
c   for use in cross-validation.
c
c   Kernel used is a rational approximation to multivariate normal, based on
c   exp(-(x^2))=1/(1 + 0.5*x^2+ 0.3*x^6); here x is the Euclidean distance
c   between data points
c
c   Inputs:
c       h               bandwidth
c       xk(n,m)         independent variables data (column=variable)
c       y(n)            dependent variable data
c       nvar            number of indep variables
c       npts            number of data points
c   Outputs:
c       f(n)            predicted values by leave-out-2W+1


      implicit double precision(a-h,o-z)
      parameter(zero=0.0,one=1.0)
      dimension xk(n,m),y(n),f(n)

      rh2= one/(h*h)
c----------- Loop on j is to predict at jth data point
      do 20 J=1,npts
        params=0.0
        tempd= 0.0
        tempn= 0.0

        do 10 I= 1,npts
          if (abs(i-j).le.jw) goto 10

c-------   loop over components of the vectors 
          dist=0.0
          do 5 jj=1,nvar
              dist= dist + (xk(j,jj)-xk(i,jj))**2
5         continue

c------     accumulate contribution from each data point
          x2=dist*rh2
          tempw= one/(one+ x2+ 0.5*x2**2 + 0.3*x2**3)
          tempn= tempn + tempw*y(i)
          tempd= tempd + tempw
10      continue

c------    normalize estimate
        f(j)= tempn/tempd

c------    accumulate contribution to effective # of parameters
        params=params+ (one/(one+tempd))

20    continue

      return    
      END    

c-------------------------------------------------------------------------
c                 Cross-validation function
c-------------------------------------------------------------------------
      function rms(h)
      implicit double precision(a-h,o-z)
      parameter(zero=0.0,one=1.0,nxmax=550,jdmax=30)
      dimension xlag(nxmax,jdmax),xt(nxmax),fhat(nxmax)
      common /xdata/ xt,xlag,cgcv,jd,nx,m,ipr,jw
c
c  Compute the leave-out-2W+1 kernel estimate at each point
c

      call kregfhat(h,nxmax,jdmax,xlag,xt,m,jd,fhat,params,jw)
c
c   subroutine kregfhat(h,n,m,xk,y,npts,nvar,f,params,jw)
c       h               bandwidth
c       xk(n,m)         independent variables data (column=variable)
c       y(n)            dependent variable data
c       npts            number of data points
c       nvar            number of indep variables
c       jw              W, for leave-out-2W+1 estimate
c   Outputs:
c       f(n)            predicted values by leave-out-2W+1
c       params          effective number of parameters, tr(A(h)) 

      sse=0.
      do 2 j=1,m
        sse=sse+(fhat(j)-xt(j))**2
2     continue
      rms=sqrt(sse/dfloat(m-1))
      if (ipr.gt.0) write(*,*) rms,h,params
c
c       Compute modified cross-validation function V_c
c
      pmax=dfloat(m)/cgcv
      if (params.ge.pmax) then
         rms=10.
      else
         xm=dfloat(m)
         rms=rms*(xm-params)/(xm-cgcv*params)
      endif
      return
      end

        function sdev(x,n1,n)
c
c    Purpose: calculates standard deviation of values n1,n2,....n
c             in array x of length n
c
        implicit double precision(a-h,o-z)
        dimension x(n)
        xn=float(n-n1+1)
        ave=0.
        do 10 i=n1,n
            ave=ave+x(i)
10      continue
        ave=ave/xn
        var=0.
        do 20 i=n1,n
            var=var+(x(i)-ave)**2
20      continue
        var=var/(xn-1.)
        sdev=sqrt(var)
        return
        end

       subroutine getx(fname,xdat,nxmax,nx,ifreq,ltr)
       implicit double precision(a-h,o-z)
       character*30 fname
       dimension xdat(nxmax)
       open(unit=11,file=fname,status='old')
       rewind(11)
       iskip=ifreq-1
       m=0
       do 50 i=1,nx
           read(11,*,end=60) xdat(i)
           m=m+1
           if (ifreq.gt.1) then
                do 40 j=1,iskip
                    read(11,*,end=60) xjunk
40              continue
           endif
           if (ltr.gt.0) xdat(i)=dlog(xdat(i))
50      continue
60     continue
       close(11)
       if (m.lt.nx) nx=m
       return
       end

        subroutine recontp(xt,xdat,xlag,nxmax,jdmax,nx,ldelay,
     *     jd,jt1,m,jtp)
        implicit double precision(a-h,o-z)
        dimension xt(nxmax),xdat(nxmax),xlag(nxmax,jdmax)
        m=0
        do 10 i=1,nx-jt1
           m=m+1
           ipred=i+jt1
           xt(i)=xdat(ipred)
           do 5 j=1,jd
               iback=jtp+(j-1)*ldelay
               xlag(i,j)=xdat(ipred-iback)
5          continue
10      continue
        return
        end

        subroutine rrecontp(xt,xdat,xlag,nxmax,jdmax,nx,ldelay,
     *     jd,jt1,m,jtp,ltr)
        implicit double precision(a-h,o-z)
        dimension xt(nxmax),xdat(nxmax),xlag(nxmax,jdmax)
        m=0
        if (ltr.eq.0) then
           do 1 i=1,nx-jt1
              m=m+1
              ipred=i+jt1
              xt(i)=dlog(xdat(ipred))-dlog(xdat(ipred-1))
1          continue
        else
           do 3 i=1,nx-jt1
              m=m+1
              ipred=i+jt1
              xt(i)=xdat(ipred)-xdat(ipred-1)
3          continue
        endif

        do 10 i=1,nx-jt1
           ipred=i+jt1
           do 5 j=1,jd
               iback=jtp+(j-1)*ldelay
               xlag(i,j)=xdat(ipred-iback)
5          continue
10      continue
        return
        end

      function ffmin(ax,bx,cx,f,tol,xmin)
      implicit double precision(a-h,o-z)
      parameter(itmax=100)
      external f
c
c      an approximation  x  to the point where  f  attains a minimum  on
c  the interval  (ax,bx)  is determined.
c
c
c  input..
c
c  ax    left endpoint of initial interval
c  cx    right endpoint of initial interval
c  f     function subprogram which evaluates  f(x)  for any  x
c        in the interval  (ax,bx)
c  tol   desired length of the interval of uncertainty of the final
c        result ( .ge. 0.0d0)
c
c
c  output..
c
c  xmin   abcissa approximating the point where f attains a minimum
c  ffmin  minimum value of f
c
c      the method used is a combination of  golden  section  search  and
c  successive parabolic interpolation.  convergence is never much slower
c  than  that  for a  fibonacci search.  if  f  has a continuous second
c  derivative which is positive at the minimum (which is not  at  ax  or
c  bx),  then  convergence  is  superlinear, and usually of the order of
c  about  1.324....
c      the function  f  is never evaluated at two points closer together
c  than  eps*abs(fmin) + (tol/3), where eps is  approximately the square
c  root  of  the  relative  machine  precision.   if   f   is a unimodal
c  function and the computed values of   f   are  always  unimodal  when
c  separated by at least  eps*abs(x) + (tol/3), then  fmin  approximates
c  the abcissa of the global minimum of  f  on the interval  ax,bx  with
c  an error less than  3*eps*abs(fmin) + tol.  if   f   is not unimodal,
c  then fmin may approximate a local, but perhaps non-global, minimum to
c  the same accuracy.
c      this function subprogram is a slightly modified  version  of  the
c  algol  60 procedure  localmin  given in richard brent, algorithms for
c  minimization without derivatives, prentice - hall, inc. (1973).
C
C     Modified by S Ellner to use relative error tolerance, tol*(1.+dabs(x)),
C     instead of absolute error tolerance; and to assume a machine
C     epsilon of 10d-20. August 1992
c
c
c  c is the squared inverse of the golden ratio
c
      c = 0.5d0*(3. - dsqrt(5.0d0))
c
c  eps is approximately the square root of the relative machine
c  precision.
c
      eps = 1.d-10
c
c  initialization
c
      a = min(ax,cx)
      b = min(ax,cx)
      v = bx
      w = v
      x = v
      e = 0.0d0
      fx = f(x)
      fv = fx
      fw = fx
c
c  main loop starts here
c
      do 86 iter=1,itmax
      xm = 0.5d0*(a + b)
      tol1 = eps + tol*(1.+dabs(x))
      tol2 = 2.0d0*tol1
c
c  check stopping criterion
c
      if (dabs(x - xm) .le. (tol2 - 0.5d0*(b - a))) go to 90
c
c is golden-section necessary
c
      if (dabs(e) .le. tol1) go to 40
c
c  fit parabola
c
      r = (x - w)*(fx - fv)
      q = (x - v)*(fx - fw)
      p = (x - v)*q - (x - w)*r
      q = 2.0d00*(q - r)
      if (q .gt. 0.0d0) p = -p
      q =  dabs(q)
      r = e
      e = d
c
c  is parabola acceptable
c
   30 if (dabs(p) .ge. dabs(0.5d0*q*r)) go to 40
      if (p .le. q*(a - x)) go to 40
      if (p .ge. q*(b - x)) go to 40
c
c  a parabolic interpolation step
c
      d = p/q
      u = x + d
c
c  f must not be evaluated too close to ax or bx
c
      if ((u - a) .lt. tol2) d = dsign(tol1, xm - x)
      if ((b - u) .lt. tol2) d = dsign(tol1, xm - x)
      go to 50
c
c  a golden-section step
c
40    if (x .ge. xm) then
          e = a - x
      else
          e = b - x
      endif
      d = c*e
c
c  f must not be evaluated too close to x
c
   50 if (dabs(d) .ge. tol1) then
          u = x + d
      else
          u = x + dsign(tol1, d)
      endif
      fu = f(u)
c
c  update  a, b, v, w, and x
c
      if (fu .gt. fx) go to 60
      if (u .ge. x) a = x
      if (u .lt. x) b = x
      v = w
      fv = fw
      w = x
      fw = fx
      x = u
      fx = fu
      go to 86
   60 if (u .lt. x) then
          a = u
      else
          b = u
      endif
      if (fu .le. fw) go to 70
      if (w .eq. x) go to 70
      if (fu .le. fv) go to 80
      if (v .eq. x) go to 80
      if (v .eq. w) go to 80
      go to 86
   70 v = w
      fv = fw
      w = u
      fw = fu
      go to 86
   80 v = u
      fv = fu
      go to 86
c
c  end of main loop
c
   86 CONTINUE
   90 xmin = x
      ffmin=fx
      return
      end
      
      subroutine  dscal(n,da,dx,incx)
c
c     scales a vector by a constant.
c     uses unrolled loops for increment equal to one.
c     jack dongarra, linpack, 3/11/78.
c
      double precision da,dx(1)
      integer i,incx,m,mp1,n,nincx
c
      if(n.le.0)return
      if(incx.eq.1)go to 20
c
c        code for increment not equal to 1
c
      nincx = n*incx
      do 10 i = 1,nincx,incx
        dx(i) = da*dx(i)
   10 continue
      return
c
c        code for increment equal to 1
c
c
c        clean-up loop
c
   20 m = mod(n,5)
      if( m .eq. 0 ) go to 40
      do 30 i = 1,m
        dx(i) = da*dx(i)
   30 continue
      if( n .lt. 5 ) return
   40 mp1 = m + 1
      do 50 i = mp1,n,5
        dx(i) = da*dx(i)
        dx(i + 1) = da*dx(i + 1)
        dx(i + 2) = da*dx(i + 2)
        dx(i + 3) = da*dx(i + 3)
        dx(i + 4) = da*dx(i + 4)
   50 continue
      return
      end

